#!/usr/bin/env python

"""
Program data structures. There are two separate kinds of structures: those with
context, which are the values manipulated by programs, and those without
context, which are typically constant things which are stored alongside the
program but which are wrapped in context-dependent structures in the running
program.

Copyright (C) 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014 Paul Boddie <paul@boddie.org.uk>

This program is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free Software
Foundation; either version 3 of the License, or (at your option) any later
version.

This program is distributed in the hope that it will be useful, but WITHOUT
ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
details.

You should have received a copy of the GNU General Public License along with
this program.  If not, see <http://www.gnu.org/licenses/>.

--------

The principal value structure class in this module is the Attr class which
represents attributes of objects and retains the context of each reference to an
attribute. This class models program behaviour at run-time.

The central data structure classes in this module are the following:

  * Class
  * Function
  * Module

All of the above support the Naming interface either explicitly or through
general conformance, meaning that all can be asked to provide their 'full_name'
using the method of that name.

Additionally, all of the above also support a dictionary interface in order to
access names within their defined scopes. Specific methods also exist in order
to distinguish between certain kinds of attributes:

  * Class: class_attributes, all_class_attributes, instance_attributes, all_attributes
  * Function: parameters, locals, all_locals
  * Module: module_attributes

These specific methods are useful in certain situations.

The above classes also provide an 'astnode' attribute, indicating the AST node
where each such object is defined.
"""

from micropython.program import ReplaceableContext, PlaceholderContext
from micropython.basicdata import *
from micropython.branch import BranchTracking
import sys

try:
    set
except NameError:
    from sets import Set as set

# Context and value manipulation.

def get_context_and_value(value):

    "Return a context, value tuple for the given 'value'."

    # Functions have a replaceable context.

    if isinstance(value, Function):
        return (ReplaceableContext, value)

    # Classes use placeholder contexts which cannot be replaced but which
    # do not communicate useful contextual information.

    elif isinstance(value, Class):
        return (PlaceholderContext, value)

    # Other values employ themselves as the context.

    else:
        return (value, value)

# Namespace classes.

class NamespaceDict(Namespace, BranchTracking):

    "A mix-in providing dictionary methods."

    def __init__(self, module=None):
        BranchTracking.__init__(self)

        self.module = module
        self.namespace = {}
        self.parent_namespaces = []
        self.globals = set()
        self.lambdas = {}           # only really useful for functions
        self.finalised = False

    def set_parent_namespaces(self, namespaces):
        self.parent_namespaces = namespaces

    def get_parent_namespaces(self):
        return self.parent_namespaces

    # Attribute/name definition and access.

    def __delitem__(self, name):
        del self.namespace[name]

    def has_key(self, name):
        return self.namespace.has_key(name)

    def keys(self):
        return self.namespace.keys()

    def values(self):
        return self.namespace.values()

    def items(self):
        return self.namespace.items()

    def __getitem__(self, name):
        return self.namespace[name]

    def get(self, name, default=None):
        return self.namespace.get(name, default)

    # Introspection methods.

    def is_method(self):

        """
        Return whether this function is a method explicitly defined in a class.
        """

        return False

    # Administrative methods.

    def finalise(self, objtable):
        self.finalise_attributes()
        self.finalise_users(objtable)

    def items_for_vacuum(self):
        return self.items() + self.lambdas.items()

    def vacuum_item(self, name):
        if self.has_key(name):
            del self[name]
            return True
        else:
            return False

    def add_lambda(self, obj):
        attr = Attr(None, self, obj.name)
        attr.update([get_context_and_value(obj)], single_assignment=1)
        self.lambdas[obj.name] = attr

    # Specialised access methods.

    def get_using_node(self, name, node):

        """
        Access the given 'name' through this namespace, making use of the module
        and builtins namespaces if necessary, annotating the given 'node' with
        the scope involved.
        """

        attr, scope, full_name = self._get_with_scope(name)

        if scope is not None:
            node._scope = scope
            self.note_scope(name, scope)

        if full_name is not None and (scope != "local" or isinstance(self, (Class, Module))):
            self.use_specific_attribute(full_name, name)

        return attr

    def get_for_local(self, name):

        """
        Get an attribute for a local name annotation, maintaining specific
        assignment information.
        """

        attr, scope, full_name = self._get_with_scope(name)

        if scope != "local":
            return None

        users = self.attribute_users[-1]
        if users.has_key(name):
            return LocalAttr(self, name, nodes=users[name], attr=attr)
        else:
            return attr

    def _get_with_scope(self, name, external=0):

        """
        Find the source of the given 'name', returning the attribute object,
        scope (constant, local, global or builtins), and the full name of the
        source namespace (or None for constants).

        If the optional 'external' flag is set to a true value, only external
        (non-local) namespaces will be involved in the search.
        """

        module = self.module
        builtins = module and module.builtins or None
        importer = module and module.importer or None

        # Constants.

        if importer and importer.predefined_constants.has_key(name):
            return importer.get_predefined_constant(name), "constant", None

        # Locals.

        if not external and self.has_key(name):
            return self[name], "local", self.full_name()

        # Outer scopes.

        for outer in self.parent_namespaces[::-1]:
            if outer.has_key(name):
                print >>sys.stderr, "Warning: name %r in namespace %r is provided by an outer scope: %r" % (name, self.full_name(), outer.full_name())
                return outer[name], "outer", outer.full_name()

        # Globals.

        if module and module is not self and module.has_key(name):
            return module[name], "global", module.full_name()

        # Builtins.

        if builtins and builtins.has_key(name):
            return builtins[name], "builtins", builtins.full_name()

        # Unknown.

        return None, None, None

    # Attribute definition methods.

    def __setitem__(self, name, value):
        self.set(name, value)

    def set(self, name, value, single_assignment=1):

        """
        A more powerful set operation, making 'name' refer to 'value' whilst
        indicating whether a 'single_assignment' (true by default) occurs in
        this operation (or whether the operation covers potentially many
        assignments in the lifetime of a program).
        """

        if value is None:
            print >>sys.stderr, "Warning: name %r in namespace %r has an unknown value (evaluated to None)." % (name, self.full_name())
            value = make_instance()

        if name in self.globals:
            self.module.set(name, value, 0)
        else:
            self._set(name, value, single_assignment)
            self.define_scope(name, "local")

    def set_module(self, name, value):

        """
        A specialised set operation, making 'name' refer to 'value' in the
        context of making a module reference available in association with
        'name' as part of the import of that module or a submodule of that
        module.
        """

        self._set(name, value, 1)

    def _set(self, name, attr_or_value, single_assignment=1):

        """
        The underlying set operation associating 'name' with the given
        'attr_or_value'.
        See: docs/assignment.txt
        """

        # For locals, direct assignments to individual name users.

        users = self.attribute_users[-1]

        if users.has_key(name):

            # Note that upon assignment there will probably be only one user.

            for user in users[name]:
                user._values = user._values or {}
                if isinstance(attr_or_value, BaseAttr):
                    attr = attr_or_value
                else:
                    attr = Attr(None, self, name)
                    self._set_using_attr(attr, attr_or_value) # enforce single assignment
                user._values[name] = attr

        # Add and/or obtain the namespace entry.

        if not self.namespace.has_key(name):
            self.namespace[name] = Attr(None, self, name)
        attr = self.namespace[name]

        # Update the attribute records.

        self._set_using_attr(attr, attr_or_value, single_assignment)

    def _set_using_attr(self, attr, attr_or_value, single_assignment=1):

        # Handle attribute assignment as well as assignment of basic objects.

        context_values = self.get_contexts_and_values(attr_or_value)
        attr.update(context_values, single_assignment)

    def get_contexts_and_values(self, attr_or_value):

        "Return adapted contexts and values for the given 'attr_or_value'."

        # Attempt to fix the context if not explicitly defined.

        if isinstance(attr_or_value, BaseAttr):
            if isinstance(attr_or_value, LocalAttr):
                attr_or_value = attr_or_value.attr
            return self.get_updated_context_values(attr_or_value.get_context_values())
        else:
            return self.get_updated_context_values([get_context_and_value(attr_or_value)])

    def get_updated_context_values(self, context_values):

        """
        Adapt the contexts found in the given 'context_values', returning a new
        set.
        See: docs/assignment.txt
        """

        results = set()

        for context, value in context_values:

            # Set the context of instances to themselves.

            if isinstance(value, Instance):
                results.add((value, value))
            else:
                results.add((context, value))

        return results

    def make_global(self, name):

        "Declare 'name' as a global in the current namespace."

        if not self.namespace.has_key(name):
            self.globals.add(name)
            self.define_scope(name, "global")
            return True
        else:
            return False

    # Attribute positioning.

    def attributes_as_list(self):

        "Return the attributes in a list."

        self.finalise_attributes()
        l = [None] * len(self.keys())
        for attr in self.values():
            l[attr.position] = attr
        return l

    def finalise_attributes(self):

        "Make sure all attributes are fully defined."

        if self.finalised:
            return

        # The default action is to assign attribute positions sequentially.

        for i, attr in enumerate(self.values()):
            attr.position = i

        self.finalised = True

    def unfinalise_attributes(self):

        "Open attribute definitions to editing and subsequent finalisation."

        self.finalised = False

# Program data structures.

def only(s):
    if len(s) == 1:
        for i in s:
            return i
    return None

class BaseAttr:

    "A basic attribute entry."

    def __init__(self, parent, name, parent_type=None):

        """
        Initialise the attribute within the collection of attributes of its
        'parent', indicating its 'name'. If the 'parent_type' is specified, it
        will contain the type of any instance parent.
        """

        self.parent = parent
        self.name = name
        self.parent_type = parent_type

        # Number of static "class" or "def" assignments per name.

        self.static_assignments = 0

    def get_contexts(self):
        return [c for (c, v) in self.get_context_values()]

    def get_values(self):
        return [v for (c, v) in self.get_context_values()]

    def get_context(self):

        "Get the context referenced by the attribute."

        if self.get_assignments() == 1:
            context_value = only(self.get_context_values())
            return context_value and context_value[0]
        else:
            return None

    def get_value(self):

        "Get the value referenced by the attribute."

        if self.get_assignments() == 1:
            context_value = only(self.get_context_values())
            return context_value and context_value[1]
        else:
            return None

    __call__ = get_value    # convenient access to any single value

    def is_constant(self):

        """
        Return whether this attribute references something that can be regarded
        as being constant within a particular scope.
        """

        return self.get_assignments() == 1

    def is_strict_constant(self):

        """
        Return whether this attribute references something that can be regarded
        as being constant.
        """

        value = self.get_value()
        return not (value is None or (isinstance(value, Instance) and not isinstance(value, Constant)))

    def is_static_attribute(self):

        """
        Return whether this attribute is defined on a fixed/static object such
        as a class or a module.
        """

        return isinstance(self.parent, (Class, Module))

    def defines_ambiguous_class(self):

        "Return whether this attribute defines more than one class."

        if self.get_assignments() > 1:
            have_class = False
            for obj in self.get_values():
                if isinstance(obj, Class):
                    if have_class:
                        return True
                    have_class = True

        return False

    def defined_within_hierarchy(self):

        """
        Return whether the parent and context of the attribute belong to the
        same class hierarchy.
        """

        # Must be defined within a class.

        if isinstance(self.parent, Class):

            # To be sure, all contexts must be classes and be the same as the
            # parent, or be a superclass of the parent, or be a subclass of the
            # parent.

            for context in self.get_contexts():
                if not (
                    isinstance(context, Class) and (
                    context is self.parent or
                    context.has_subclass(self.parent) or
                    self.parent.has_subclass(context))
                    ):
                    return False

            return True

        # Instance attributes are not defined within a hierarchy.

        else:
            return False

    def defined_outside_hierarchy(self):

        """
        Return whether the parent and context of the attribute never belong to
        the same class hierarchy.
        """

        # Must be defined within a class.

        if isinstance(self.parent, Class):

            # To be sure, all contexts must be classes and be the same as the
            # parent, or be a superclass of the parent, or be a subclass of the
            # parent.

            for context in self.get_contexts():
                if (
                    isinstance(context, Class) and not (
                    context is self.parent or
                    context.has_subclass(self.parent) or
                    self.parent.has_subclass(context))
                    ):
                    return False

            return True

        # Instance attributes are not defined within a hierarchy.

        else:
            return True

    def __repr__(self):
        return "<attribute %s.%s (%sassigned %r)>" % (
            shortrepr(self.parent), self.name,
            self._repr_parent_type(), self.get_assignments()
            )

    def _repr_parent_type(self):
        if self.parent_type is not None:
            return "parent %s, " % shortrepr(self.parent_type)
        else:
            return ""

    def __shortrepr__(self):
        return "%s.%s" % (shortrepr(self.parent), self.name)

class LocalAttr(BaseAttr):

    """
    An attribute of a local namespace - a local name - derived from potentially
    many assignments. This attribute dynamically generates value information
    from that stored on assignment nodes so that it may be possible to update
    such nodes and thus the knowledge about the nature of the attribute at a
    later point.
    """

    def __init__(self, parent, name, nodes, attr):
        BaseAttr.__init__(self, parent, name)
        self.nodes = nodes or set()
        self.attr = attr
        self.users = None

    def _get_defining_users(self):
        if self.users is None:
            users = set()
            for node in self.nodes:
                for user in node._attrdefs or [node]:
                    if user._values and user._values.has_key(self.name):
                        users.add(user)
            self.users = users
        return self.users

    def get_assignments(self):
        return len(self._get_defining_users())

    def get_context_values(self):

        """
        Return the context, value pairs for the attribute. Since these may be
        derived from value information obtained from the defining users, and
        since such information may be changed during the deduction process, the
        result of this method may change after initial inspection.
        """

        context_values = set()
        for def_user in self._get_defining_users():
            attr = def_user._values[self.name]
            if isinstance(attr, BaseAttr):
                context_values.update(attr.get_context_values())
            else:
                context_values.add(get_context_and_value(attr))
        return context_values

class Attr(BaseAttr):

    "An attribute entry having context and value information."

    def __init__(self, position, parent, name, parent_type=None):

        """
        Initialise the attribute with the given 'position' within the collection
        of attributes of its 'parent', indicating its 'name'. If the
        'parent_type' is specified, it will contain the type of any instance
        parent.
        """

        BaseAttr.__init__(self, parent, name, parent_type)
        self.position = position

        # Possible values.

        self.context_values = set()

        # Number of assignments per name.

        self.assignments = None

    def get_assignments(self):
        return self.assignments

    def get_context_values(self):
        return self.context_values

    # Generic update operations on attributes.

    def update(self, context_values, single_assignment):

        """
        Update the attribute, adding the 'context_values' provided to the
        known details associated with the attribute, changing the number of
        assignments according to the 'single_assignment' status of the
        operation, where a true value indicates that only one assignment is
        associated with the update, and a false value indicates that potentially
        many assignments may be involved.
        """

        if self.context_values.issuperset(context_values) and \
            not (make_instance(), make_instance()) in context_values:
            return

        self.update_assignments(len(set(context_values)), single_assignment)
        self.context_values.update(context_values)

    def update_assignments(self, n, single_assignment):

        """
        Update the assignment count by 'n' or "at least" 'n' if
        'single_assignment' is given as a false value.
        """

        if self.assignments is None:
            if single_assignment:
                self.assignments = n
            else:
                self.assignments = AtLeast(n)
        else:
            if single_assignment:
                self.assignments += n
            else:
                self.assignments += AtLeast(n)

    def __repr__(self):
        return "<attribute %s.%s (%s%sassigned %r)>" % (
            shortrepr(self.parent), self.name,
            self._repr_parent_type(), self._repr_position(),
            self.get_assignments()
            )

    def _repr_position(self):
        if self.position is not None:
            return "at %r, " % self.position
        else:
            return ""

    def __shortrepr__(self):
        return "%s.%s (at %r)" % (shortrepr(self.parent), self.name,
            self._repr_position())

class Class(NamespaceDict, Naming, Constant):

    "A base class for common/normal classes and the type class."

    def __init__(self, name, parent=None, module=None, node=None, original_name=None):

        """
        Initialise the class with the given 'name', optional 'parent' object,
        'module' and AST 'node'. The optional information must be set at a later
        point using the 'set_context' method if omitted.
        """

        NamespaceDict.__init__(self, module)
        self.name = name
        self.parent = parent
        self.astnode = node
        self.original_name = original_name or name

        # Superclasses, descendants and attributes.

        self.bases = []
        self.descendants = set()
        self.instattr = set()           # instance attributes
        self.instattr_tentative = set() # tentative/suspected instance attributes
        self.relocated = set()          # attributes which do not have the same
                                        # position as those of the same name in
                                        # some superclasses

        # Caches.

        self.reset_caches()

        # Program-related details.

        self.instantiator = None
        self.temp_usage = 0
        self.local_usage = 0
        self.all_local_usage = 0

        # Add an attribute to this class for use by instances.

        self.set("__class__", self)

    def set_context(self, parent, module, node):

        "Set the 'parent', 'module' and 'node' of a class created in advance."

        self.parent = parent
        self.module = module
        self.astnode = node

    def reset_caches(self):

        "Reset the caches."

        self.all_instattr = None        # cache for instance_attributes
        self.all_instattr_names = None  # from all_instattr
        self.all_classattr = None       # cache for all_class_attributes
        self.all_classattr_names = None # from all_classattr
        self.allattr = None             # cache for all_attributes
        self.allattr_names = None       # from allattr

    def __repr__(self):
        return "<class %s>" % shortrepr(self)

    def __shortrepr__(self):
        return "%s.%s" % (shortrepr(self.parent), self.name)

    # Namespace-related methods.

    def get_updated_context_values(self, context_values):

        """
        Adapt the contexts found in the given 'context_values', returning a new
        set.
        See: docs/assignment.txt
        """

        results = set()

        for context, value in context_values:

            # Change the ownership of functions.

            if context is ReplaceableContext and value is not None and isinstance(value, Function):
                results.add((self, value))
            else:
                results.add((context, value))

        return NamespaceDict.get_updated_context_values(self, results)

    # Administrative methods.

    def items_for_vacuum(self):

        "Consider both class and instance attributes for vacuuming."

        items = []
        for name in self.instattr:
            items.append((name, None))
        return NamespaceDict.items_for_vacuum(self) + items

    def vacuum_item(self, name):

        "Vacuum 'name' from the class or instance attribute collections."

        # NOTE: Hack to prevent damage to exceptions.

        if name == "_pc":
            return False

        if not NamespaceDict.vacuum_item(self, name):
            self.instattr.remove(name)
        return True

    def finalise_attributes(self):

        "Make sure that all attributes are fully defined."

        if self.finalised:
            return

        self.finalise_class_attributes()
        self.finalise_instance_attributes()
        self.finalised = True

    def unfinalise_attributes(self):

        "Open attribute definitions to editing and subsequent finalisation."

        self.reset_caches()
        self.finalised = False

    # Convenience methods for accessing functions and methods.

    def get_instantiator(self):

        "Return a function which can be used to instantiate the class."

        if self.instantiator is None:
            self.instantiator = self.get_init_method().as_instantiator()
        return self.instantiator

    def get_init_method(self):
        return self.all_class_attributes()["__init__"].get_value()

    # Class-specific methods.

    def add_base(self, base):
        self.bases.append(base)
        base.add_descendant(self)

    def add_instance_attribute(self, name, tentative=False):
        if tentative:
            self.instattr_tentative.add(name)
        else:
            self.instattr.add(name)

    def add_descendant(self, cls):
        self.descendants.add(cls)
        for base in self.bases:
            base.add_descendant(cls)

    def has_subclass(self, other):
        return other in self.descendants

    def all_descendants(self):
        d = {}
        for cls in self.descendants:
            d[cls.full_name()] = cls
        return d

    "Return the attribute names provided by this class only."

    class_attribute_names = NamespaceDict.keys

    def class_attributes(self):

        "Return class attributes provided by this class only."

        return dict(self)

    def all_class_attribute_names(self):

        "Return the attribute names provided by classes in this hierarchy."

        if self.all_classattr_names is None:
            self.all_class_attributes()
            self.all_classattr_names = self.all_classattr.keys()
        return self.all_classattr_names

    def all_class_attributes(self):

        "Return all class attributes, indicating the class which provides them."

        self.finalise_class_attributes()
        return self.all_classattr

    def finalise_class_attributes(self):

        "Make sure that the class attributes are fully defined."

        if self.all_classattr is None:
            self.all_classattr = {}
            clsattr = {}

            # Record provisional position information for attributes of this
            # class.

            for name in self.class_attributes().keys():

                # Special case: __class__ has to be at position 0.

                if name == "__class__":
                    clsattr[name] = set([0])
                else:
                    clsattr[name] = set() # position not yet defined

            reversed_bases = self.bases[:]
            reversed_bases.reverse()

            # For the bases in reverse order, acquire class attribute details.

            for cls in reversed_bases:
                for name, attr in cls.all_class_attributes().items():
                    self.all_classattr[name] = attr

                    # Record previous attribute information.

                    if clsattr.has_key(name):
                        clsattr[name].add(attr.position)

            # Record class attributes provided by this class and its bases,
            # along with their positions.

            self.all_classattr.update(self.class_attributes())

            if clsattr:
                for i, name in enumerate(self._get_position_list(clsattr)):
                    self.all_classattr[name].position = i

        return self.all_classattr

    def instance_attribute_names(self):

        "Return the instance attribute names provided by the class."

        if self.all_instattr_names is None:
            self.instance_attributes()
        return self.all_instattr_names

    def instance_attributes(self):

        "Return instance-only attributes for instances of this class."

        self.finalise_instance_attributes()
        return self.all_instattr

    def instance_attributes_as_list(self):

        "Return instance-only attributes in a list ordered by position."

        attrs = self.instance_attributes().values()
        attrs.sort(cmp=lambda x, y: cmp(x.position, y.position))
        return attrs

    def finalise_instance_attributes(self):

        "Make sure that the instance attributes are fully defined."

        # Eliminate tentative instance attributes that are actually class
        # attributes.

        for attrname in self.all_class_attributes().keys():
            if attrname in self.instattr_tentative:
                self.instattr_tentative.remove(attrname)

        for cls in self.descendants:
            for attrname in cls.class_attribute_names():
                if attrname in self.instattr_tentative:
                    self.instattr_tentative.remove(attrname)

        for attrname in self.instattr_tentative:
            self.instattr.add(attrname)

        # Cache the attributes by converting the positioned attributes into a
        # dictionary.

        if self.all_instattr is None:
            self.all_instattr = self._get_attributes()
            self.all_instattr_names = self.all_instattr.keys()

        return self.all_instattr

    def _get_attributes(self):

        """
        Return a dictionary mapping names to Attr instances incorporating
        information about their positions in the final instance structure.
        """

        instattr = {}

        # Record provisional position information for attributes of this
        # instance.

        for name in self.instattr:
            instattr[name] = set() # position not yet defined

        reversed_bases = self.bases[:]
        reversed_bases.reverse()

        # For the bases in reverse order, acquire instance attribute
        # details.

        for cls in reversed_bases:
            for name, attr in cls.instance_attributes().items():

                # Record previous attribute information.

                if instattr.has_key(name):
                    instattr[name].add(attr.position)
                else:
                    instattr[name] = set([attr.position])

        # Build the dictionary of attributes using the existing positions known
        # for each name.

        d = {}
        for i, name in enumerate(self._get_position_list(instattr)):
            d[name] = Attr(i, make_instance(), name, self)
        return d

    def _get_position_list(self, positions):

        """
        Return a list of attribute names for the given 'positions' mapping from
        names to positions, indicating the positions of the attributes in the
        final instance structure.
        """

        position_items = positions.items()
        namearray = [None] * len(position_items)

        # Get the positions in ascending order of list size, with lists
        # of the same size ordered according to their smallest position
        # value.

        position_items.sort(self._cmp_positions)

        # Get the names in position order.

        held = []

        for name, pos in position_items:
            pos = list(pos)
            pos.sort()
            if pos and pos[0] < len(namearray) and namearray[pos[0]] is None:
                namearray[pos[0]] = name
            else:
                if pos:
                    self.relocated.add(name)
                held.append((name, pos))

        for i, attr in enumerate(namearray):
            if attr is None:
                name, pos = held.pop()
                namearray[i] = name

        return namearray

    def _cmp_positions(self, a, b):

        "Compare name plus position list operands 'a' and 'b'."

        name_a, list_a = a
        name_b, list_b = b
        if len(list_a) < len(list_b):
            return -1
        elif len(list_a) > len(list_b):
            return 1
        elif not list_a:
            return 0
        else:
            return cmp(min(list_a), min(list_b))

    def all_attribute_names(self):

        """
        Return the names of all attributes provided by instances of this class.
        """

        self.allattr_names = self.allattr_names or self.all_attributes().keys()
        return self.allattr_names

    def all_attributes(self):

        """
        Return all attributes for an instance, indicating either the class which
        provides them or that the instance itself provides them.

        Note that __class__ acts like a class attribute for both instances and
        classes, and must be able to convey distinct values.
        """

        if self.allattr is None:
            self.allattr = {}
            self.allattr.update(self.all_class_attributes())
            for name, attr in self.instance_attributes().items():
                if self.allattr.has_key(name) and name != "__class__":
                    print >>sys.stderr, "Warning: instance attribute %r in %r overrides class attribute." % (name, self)
                self.allattr[name] = attr
        return self.allattr

class Function(NamespaceDict, Naming, Constant):

    "An inspected function."

    def __init__(self, name, parent, argnames, defaults, has_star, has_dstar,
        dynamic_def=0, module=None, node=None, original_name=None):

        """
        Initialise the function with the given 'name', 'parent', list of
        'argnames', list of 'defaults', the 'has_star' flag (indicating the
        presence of a * parameter), the 'has_dstar' flag (indicating the
        presence of a ** parameter), optional 'dynamic_def' (indicating that the
        function must be handled dynamically), optional 'module', and optional
        AST 'node'.
        """

        NamespaceDict.__init__(self, module)

        if name is None:
            self.name = "lambda#%d" % new_lambda()
            self._is_lambda = True
        else:
            self.name = name
            self._is_lambda = False

        self.original_name = original_name or name

        self.parent = parent
        self.argnames = argnames
        self.defaults = defaults
        self.has_star = has_star
        self.has_dstar = has_dstar
        self.dynamic_def = dynamic_def
        self.astnode = node

        # Initialise the positional names.

        self.positional_names = self.argnames[:]
        if has_dstar:
            self.dstar_name = self.positional_names[-1]
            del self.positional_names[-1]
        if has_star:
            self.star_name = self.positional_names[-1]
            del self.positional_names[-1]

        # Initialise default storage.
        # NOTE: This must be initialised separately due to the reliance on node
        # NOTE: visiting.

        self.default_attrs = []

        # Initialise attribute usage.

        if node is not None:
            for arg in argnames:

                # Define attribute users.

                self._define_attribute_user_for_name(node, arg)

        # Caches.

        self.localnames = None # cache for locals

        # Add parameters to the namespace.

        self._add_parameters(argnames)

        # Image generation details.

        self.dynamic = None

        # Program-related details.

        self.temp_usage = 0
        self.local_usage = 0
        self.all_local_usage = 0

    def _add_parameters(self, argnames):

        "Add 'argnames' to the namespace."

        for name in argnames:
            self.set(name, make_instance())

        for name, top_level in self._flattened_parameters(argnames):
            if not top_level:
                self.set(name, make_instance())

    def _flattened_parameters(self, argnames, top_level=1):
        l = []
        for name in argnames:
            if isinstance(name, tuple):
                l += self._flattened_parameters(name, 0)
            else:
                l.append((name, top_level))
        return l

    def __repr__(self):
        return "<function %s>" % shortrepr(self)

    def __shortrepr__(self):
        return "%s.%s(%s)" % (shortrepr(self.parent), self.name, ", ".join(self.argnames))

    def is_lambda(self):
        return self._is_lambda

    # Defaults-related methods.

    def store_default(self, attr_or_value):

        """
        Reserve space for defaults, set outside the function, potentially on a
        dynamic basis, using the 'attr_or_value'.
        """

        attr = Attr(None, self, None)
        self._set_using_attr(attr, attr_or_value)
        self.default_attrs.append(attr)

    def make_dynamic(self):

        "Return whether this function must be handled using a dynamic object."

        if self.dynamic is None:
            for attr in self.default_attrs:
                if not attr.is_strict_constant() and self.dynamic_def:
                    self.dynamic = True
                    self._make_dynamic()
                    break
            else:
                self.dynamic = False

        return self.dynamic

    is_dynamic = make_dynamic

    def _make_dynamic(self):

        "Where functions have dynamic defaults, add a context local."

        self.set("__context__", make_instance())

    # Namespace-related methods.

    def make_global(self, name):

        "Declare 'name' as a global in the current namespace."

        if name not in self.argnames and not self.has_key(name):
            self.globals.add(name)
            return True
        else:
            return False

    def parameters(self):

        """
        Return a dictionary mapping parameter names to their position in the
        parameter list.
        """

        parameters = {}
        for i, name in enumerate(self.argnames):
            parameters[name] = i
        return parameters

    def tuple_parameters(self, argnames=None):

        """
        Return a list of (position, parameter) entries corresponding to tuple
        parameters, where each parameter may either be a string or another such
        list of entries.
        """

        names = argnames or self.argnames

        l = []
        for i, name in enumerate(names):
            if isinstance(name, tuple):
                l.append((i, self.tuple_parameters(name)))
            elif argnames:
                l.append((i, name))
        return l

    def all_locals(self):

        "Return a dictionary mapping names to local and parameter details."

        return dict(self)

    def locals(self):

        "Return a dictionary mapping names to local details."

        if self.localnames is None:
            self.localnames = {}
            self.localnames.update(self.all_locals())
            for name in self.argnames:
                del self.localnames[name]
        return self.localnames

    def is_method(self):

        """
        Return whether this function is a method explicitly defined in a class.
        """

        return isinstance(self.parent, Class)

    def is_relocated(self, name):

        """
        Determine whether the given attribute 'name' is relocated for instances
        having this function as a method.
        """

        for cls in self.parent.descendants:
            if name in cls.relocated:
                return True
        return False

    # Administrative methods.

    def items_for_vacuum(self):
        return self.lambdas.items()

    def vacuum_item(self, name):
        del self.lambdas[name]
        return True

    def finalise_attributes(self):

        """
        Make sure all attributes (locals) are fully defined. Note that locals
        are not attributes in the sense of class, module or instance attributes.
        Defaults are also finalised by this method.
        """

        if self.finalised:
            return

        # Defaults are positioned in the function structure.

        for i, default in enumerate(self.default_attrs):
            default.position = i

        # Parameters.

        i = self._finalise_parameters()

        if i is not None:
            nparams = i + 1
        else:
            nparams = 0

        # Locals (and tuple parameter names) are positioned in the current stack
        # frame.

        i = None
        for i, attr in enumerate(self.locals().values()):
            attr.position = i + nparams

        if i is not None:
            nothers = i + 1
        else:
            nothers = 0

        self.local_usage = nothers
        self.all_local_usage = nparams + nothers
        self.finalised = True

    def _finalise_parameters(self):
        if not self.argnames:
            return None

        for i, name in enumerate(self.argnames):
            self[name].position = i

        return i

    def as_instantiator(self):

        "Make an instantiator function from a method, keeping all arguments."

        function = Function(self.parent.name, self.parent.parent, self.argnames, self.defaults,
            self.has_star, self.has_dstar, self.dynamic_def, self.module)
        function.default_attrs = self.default_attrs
        return function

class UnresolvedName(NamespaceDict, Constant):

    "A module, class or function which was mentioned but could not be imported."

    def __init__(self, name, parent_name, module=None):
        NamespaceDict.__init__(self, module)
        self.name = name
        self.parent_name = parent_name
        self.parent = None

        self.descendants = set()

    def add_descendant(self, cls):
        self.descendants.add(cls)

    def all_attributes(self):
        return {}

    def all_attribute_names(self):
        return []

    all_class_attributes = class_attributes = instance_attributes = all_attributes
    all_class_attribute_names = class_attribute_names = instance_attribute_names = all_attribute_names

    def __repr__(self):
        return "<unknown %s>" % shortrepr(self)

    def __shortrepr__(self):
        if self.name is not None:
            return "%s.%s" % (self.parent_name, self.name)
        else:
            return self.parent_name

    def full_name(self):
        if self.name is not None:
            return self.parent_name + "." + self.name
        else:
            return self.parent_name

class Module(NamespaceDict, Constant):

    "An inspected module's core details."

    def __init__(self, name, importer):
        NamespaceDict.__init__(self, self)
        self.name = name
        self.importer = importer
        self.parent = None

        # Original location details.

        self.astnode = None

        # Complete lists of classes and functions.

        self.all_objects = set()

        # Whether the module is imported in a circular fashion, exposing the
        # unfinished namespace to external modification.

        self.circular_import = self in importer.circular_imports

        # A set of global names that cannot support combined attribute usage
        # observations because they may be modified within functions during
        # initialisation.

        self.modified_names = set()

        # Keyword records.

        self.keyword_names = set()

        # Program-related details.

        self.temp_usage = 0
        self.local_usage = 0
        self.all_local_usage = 0

    def full_name(self):
        return self.name

    def __repr__(self):
        return "<module %s>" % shortrepr(self)

    def __shortrepr__(self):
        return self.name

    # Attribute methods.

    "Return the module attribute names provided by the module."

    module_attribute_names = NamespaceDict.keys

    def module_attributes(self):

        "Return a dictionary mapping names to module attributes."

        return dict(self)

    all_attributes = module_attributes

    def get_static_attribute(self, name):

        """
        Return a static attribute for the given 'name' or None if no such
        attribute exists.
        """

        return self.get(name)

    def modify_name(self, name):

        """
        Modify a global 'name' invalidating various assumptions about its
        behaviour based on the module namespace being "safe" and suitable for
        attribute usage and constant value observations.
        """

        self.modified_names.add(name)

        # Establish an attribute directly in the namespace if not present.

        if not self.namespace.has_key(name):
            self.namespace[name] = Attr(None, self, name)

        # Indicate that there is at least one assignment to the name, although
        # no actual values are recorded.

        attr = self.namespace[name]
        attr.update_assignments(1, False)

    def set(self, name, value, single_assignment=1):

        """
        Set the module attribute with the given 'name', having the given 'value'
        that is assigned once if 'single_assignment' is omitted to given as a
        true value.

        Where the module is imported before it is completely initialised, all
        assignments will be regarded as multiple assignments since it is
        possible that an external assignment to the name may occur.
        """

        NamespaceDict.set(self, name, value, not self.circular_import and single_assignment)

    # Attribute usage methods that apply to module globals in certain
    # circumstances.

    def can_use_name_for_usage(self, name):
        return name not in self.modified_names and not self.circular_import

    def _use_attribute(self, name, attrname, value=None):
        if self.can_use_name_for_usage(name):
            return NamespaceDict._use_attribute(self, name, attrname, value)
        else:
            self.importer.use_name(attrname, self.full_name(), value)
            return []

    def _define_attribute_user_for_name(self, node, name):
        if self.can_use_name_for_usage(name):
            NamespaceDict._define_attribute_user_for_name(self, node, name)

    def _init_attribute_user_for_name(self, node, name):
        if self.can_use_name_for_usage(name):
            NamespaceDict._init_attribute_user_for_name(self, node, name)

    def _define_attribute_accessor(self, name, attrname, node, value):
        if self.can_use_name_for_usage(name):
            NamespaceDict._define_attribute_accessor(self, name, attrname, node, value)
        else:
            self.importer.use_name(attrname, self.full_name(), value)

# Pre-made class instances.
# For each of these, their details will be filled in later.

premade = {
    "bool" : Class("bool"),
    "float" : Class("float"),
    "int" : Class("int"),
    "long" : Class("long"),
    "str" : Class("str"),
    "unicode" : Class("unicode"),
    "type" : Class("type"),
    "NoneType" : Class("NoneType"),
    }

def get_constant_class(name):
    return premade[name]

# Class construction.

def get_class(name, parent, module, node):

    """
    Return a Class instance for the class with the given 'name', 'parent',
    'module' and 'node'.
    """

    if premade.has_key(name) and module.full_name() == "__builtins__":
        cls = premade[name]
        cls.set_context(parent, module, node)
    else:
        # Where names are reused in a namespace, differentiate between classes
        # using a name index.

        original_name = name

        if parent.has_key(name):
            assignments = parent[name].static_assignments
            if assignments >= 1:
                name = "%s#%d" % (name, assignments + 1)

        cls = Class(name, parent, module, node, original_name)

        # Add a reference for the class's "shadow" name.

        parent.use_specific_attribute(parent.full_name(), name)

    return cls

# Function construction.

def get_function(name, parent, argnames, defaults, has_star, has_dstar,
    dynamic_def=0, module=None, node=None):

    """
    Return a Function instance for the class with the given 'name', 'parent',
    and other details.
    """

    original_name = name

    if parent.has_key(name):
        assignments = parent[name].static_assignments
        if assignments >= 1:
            name = "%s#%d" % (name, assignments + 1)

    fn = Function(name, parent, argnames, defaults, has_star, has_dstar,
        dynamic_def, module, node, original_name)

    # Add a reference for the function's "shadow" name.

    if name:
        parent.use_specific_attribute(parent.full_name(), name)

    return fn

# Lambda sequence numbering.

lambda_index = 0

def new_lambda():

    "Return a new sequence number for a lambda definition."

    global lambda_index
    lambda_index += 1
    return lambda_index

# Special representations.

class AtLeast:

    "A special representation for numbers of a given value or greater."

    def __init__(self, count):
        self.count = count

    def __eq__(self, other):
        return False

    __lt__ = __le__ = __eq__

    def __ne__(self, other):
        return True

    def __gt__(self, other):
        if isinstance(other, AtLeast):
            return False
        else:
            return self.count > other

    def __ge__(self, other):
        if isinstance(other, AtLeast):
            return False
        else:
            return self.count >= other

    def __iadd__(self, other):
        if isinstance(other, AtLeast):
            self.count += other.count
        else:
            self.count += other
        return self

    def __radd__(self, other):
        if isinstance(other, AtLeast):
            return AtLeast(self.count + other.count)
        else:
            return AtLeast(self.count + other)

    def __repr__(self):
        return "AtLeast(%r)" % self.count

# vim: tabstop=4 expandtab shiftwidth=4
